﻿' demo of multi-swarm optimization

Imports System
Namespace MultiSwarm
    Friend Class MultiSwarmProgram
        Shared Sub Main(ByVal args() As String)
            Try
                Console.WriteLine(vbCrLf & "Begin Multiple Particle Swarm optimization demo" & vbCrLf)
                Console.WriteLine("Goal is to minimize Rastrigin's function")
                Console.WriteLine("Function has known solution of 0.0 at x0 = 0.0, x1 = 0.0")

                Console.WriteLine(vbCrLf & "Setting minX = -100.0 and maxX = +100.0")
                Dim [dim] As Integer = 2
                Dim minX As Double = -100.0
                Dim maxX As Double = 100.0

                Console.WriteLine("Setting number particles in each swarm = 4")
                Console.WriteLine("Setting number swarms in multiswarm = 3")
                Dim numParticles As Integer = 4 ' number particles in each swarm
                Dim numSwarms As Integer = 3 ' number swarms in multiswarm

                Console.WriteLine(vbCrLf & "Initializing all swarms in multiswarm")
                Dim ms As New Multiswarm(numSwarms, numParticles, [dim], minX, maxX)
                Console.WriteLine(vbCrLf & "Initial multiswarm:")
                Console.WriteLine(ms.ToString)

                Dim maxLoop As Integer = 150
                Console.WriteLine(vbCrLf & "Setting maxLoop = " & maxLoop)
                Console.WriteLine("Entering main solve loop")
                ms.Solve(maxLoop)

                Console.WriteLine("Solve loop complete")
                Console.WriteLine(vbCrLf & "Final multiswarm:")
                Console.WriteLine(ms.ToString)

                Console.WriteLine(vbCrLf & "Best solution found = " & ms.bestMultiCost.ToString("F6"))
                Console.Write("at x0 = " & ms.bestMultiPos(0).ToString("F4"))
                Console.WriteLine(", x1 = " & ms.bestMultiPos(1).ToString("F4"))

                Console.WriteLine(vbCrLf & "End demo" & vbCrLf)
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine(ex.Message)
                Console.ReadLine()
            End Try
        End Sub

        Public Shared Function Cost(ByVal position() As Double) As Double
            Dim result As Double = 0.0
            For i = 0 To position.Length - 1
                Dim xi As Double = position(i)
                result += (xi * xi) - (10 * Math.Cos(2 * Math.PI * xi)) + 10
            Next i
            Return result
        End Function

    End Class ' Program

    Public Class Particle
        Private Shared ran As New Random(0)
        Public position() As Double
        Public velocity() As Double
        Public cost As Double
        Public bestPartPos() As Double
        Public bestPartCost As Double

        Public Sub New(ByVal [dim] As Integer, ByVal minX As Double, ByVal maxX As Double)
            position = New Double([dim] - 1) {}
            velocity = New Double([dim] - 1) {}
            bestPartPos = New Double([dim] - 1) {}
            For i = 0 To [dim] - 1
                position(i) = (maxX - minX) * ran.NextDouble + minX
                velocity(i) = (maxX - minX) * ran.NextDouble + minX
            Next i
            cost = MultiSwarmProgram.Cost(position)
            bestPartCost = cost
            Array.Copy(position, bestPartPos, [dim])
        End Sub

        Public Overrides Function ToString() As String
            Dim s As String = ""
            s &= "Pos [ "
            For i = 0 To position.Length - 1
                s &= position(i).ToString("F2") & " "
            Next i
            s &= "] "
            s &= "Vel [ "
            For i = 0 To velocity.Length - 1
                s &= velocity(i).ToString("F2") & " "
            Next i
            s &= "] "
            s &= "Cost = " & cost.ToString("F3")
            s &= " Best Pos [ "
            For i = 0 To bestPartPos.Length - 1
                s &= bestPartPos(i).ToString("F2") & " "
            Next i
            s &= "] "
            s &= "BestCost = " & cost.ToString("F3")
            Return s
        End Function

    End Class ' Particle

    Public Class Swarm
        Public particles() As Particle
        Public bestSwarmPos() As Double
        Public bestSwarmCost As Double

        Public Sub New(ByVal numParticles As Integer, ByVal [dim] As Integer, ByVal minX As Double, ByVal maxX As Double)
            bestSwarmCost = Double.MaxValue
            bestSwarmPos = New Double([dim] - 1) {}
            particles = New Particle(numParticles - 1) {}
            For i = 0 To particles.Length - 1
                particles(i) = New Particle([dim], minX, maxX)
                If particles(i).cost < bestSwarmCost Then
                    bestSwarmCost = particles(i).cost
                    Array.Copy(particles(i).position, bestSwarmPos, [dim])
                End If
            Next i
        End Sub

        Public Overrides Function ToString() As String
            Dim s As String = ""
            For i = 0 To particles.Length - 1
                s &= "[" & i & "] " & particles(i).ToString & vbCrLf
            Next i
            s &= "Best Swarm Pos [ "
            For i = 0 To bestSwarmPos.Length - 1
                s &= bestSwarmPos(i).ToString("F2") & " "
            Next i
            s &= "] "
            s &= "Best Swarm Cost = " & bestSwarmCost.ToString("F3")
            s &= vbCrLf
            Return s
        End Function
    End Class ' Swarm

    Public Class Multiswarm
        Public swarms() As Swarm
        Public bestMultiPos() As Double
        Public bestMultiCost As Double
        Public [dim] As Integer
        Public minX As Double
        Public maxX As Double

        Private Shared ran As New Random(0) ' Solve() has random features

        Public Sub New(ByVal numSwarms As Integer, ByVal numParticles As Integer, ByVal [dim] As Integer, ByVal minX As Double, ByVal maxX As Double)
            swarms = New Swarm(numSwarms - 1) {}
            bestMultiPos = New Double([dim] - 1) {}
            bestMultiCost = Double.MaxValue
            Me.dim = [dim]
            Me.minX = minX
            Me.maxX = maxX

            For i = 0 To numSwarms - 1
                swarms(i) = New Swarm(numParticles, [dim], minX, maxX)
                If swarms(i).bestSwarmCost < bestMultiCost Then
                    bestMultiCost = swarms(i).bestSwarmCost
                    Array.Copy(swarms(i).bestSwarmPos, bestMultiPos, [dim])
                End If
            Next i
        End Sub

        Public Sub Solve(ByVal maxLoop As Integer)
            Dim ct As Integer = 0
            Dim w As Double = 0.729 ' inertia
            Dim c1 As Double = 1.49445 ' particle / cogntive
            Dim c2 As Double = 1.49445 ' swarm / social
            Dim c3 As Double = 0.3645 ' multiswarm / global
            Dim death As Double = 0.005 ' prob of particle death

            Dim immigrate As Double = 0.005 ' prob of particle immigration

            Do While ct < maxLoop
                ct += 1
                For i = 0 To swarms.Length - 1 ' each swarm
                    For j = 0 To swarms(i).particles.Length - 1 ' each particle
                        Dim p As Double = ran.NextDouble
                        If p < death Then
                            swarms(i).particles(j) = New Particle([dim], minX, maxX)
                        End If

                        Dim q As Double = ran.NextDouble
                        If q < immigrate Then
                            Immigration(i, j) ' swap curr particle with a random particle in diff swarm
                        End If

                        For k = 0 To [dim] - 1 ' update velocity. each x position component
                            Dim r1 As Double = ran.NextDouble
                            Dim r2 As Double = ran.NextDouble
                            Dim r3 As Double = ran.NextDouble

                            swarms(i).particles(j).velocity(k) = (w * swarms(i).particles(j).velocity(k)) + (c1 * r1 * (swarms(i).particles(j).bestPartPos(k) - swarms(i).particles(j).position(k))) + (c2 * r2 * (swarms(i).bestSwarmPos(k) - swarms(i).particles(j).position(k))) + (c3 * r3 * (bestMultiPos(k) - swarms(i).particles(j).position(k)))

                            If swarms(i).particles(j).velocity(k) < minX Then
                                swarms(i).particles(j).velocity(k) = minX
                            ElseIf swarms(i).particles(j).velocity(k) > maxX Then
                                swarms(i).particles(j).velocity(k) = maxX
                            End If

                        Next k

                        For k = 0 To [dim] - 1 ' update position
                            swarms(i).particles(j).position(k) += swarms(i).particles(j).velocity(k)
                        Next k

                        ' update cost
                        swarms(i).particles(j).cost = MultiSwarmProgram.Cost(swarms(i).particles(j).position)

                        ' check if new best cost
                        If swarms(i).particles(j).cost < swarms(i).particles(j).bestPartCost Then
                            swarms(i).particles(j).bestPartCost = swarms(i).particles(j).cost
                            Array.Copy(swarms(i).particles(j).position, swarms(i).particles(j).bestPartPos, [dim])
                        End If

                        If swarms(i).particles(j).cost < swarms(i).bestSwarmCost Then
                            swarms(i).bestSwarmCost = swarms(i).particles(j).cost
                            Array.Copy(swarms(i).particles(j).position, swarms(i).bestSwarmPos, [dim])
                        End If

                        If swarms(i).particles(j).cost < bestMultiCost Then
                            bestMultiCost = swarms(i).particles(j).cost
                            Array.Copy(swarms(i).particles(j).position, bestMultiPos, [dim])
                        End If
                    Next j
                Next i

            Loop
        End Sub

        Private Sub Immigration(ByVal i As Integer, ByVal j As Integer)
            ' swap particle j in swarm i, with a random particle in a random swarm
            Dim otheri As Integer = ran.Next(0, swarms.Length)
            Dim otherj As Integer = ran.Next(0, swarms(0).particles.Length)
            Dim tmp As Particle = swarms(i).particles(j)
            swarms(i).particles(j) = swarms(otheri).particles(otherj)
            swarms(otheri).particles(otherj) = tmp
        End Sub

        Public Overrides Function ToString() As String
            Dim s As String = ""
            s &= "=======================" & vbCrLf
            For i = 0 To swarms.Length - 1
                s &= swarms(i).ToString & vbCrLf
            Next i
            s &= "Best Multiswarm Pos [ "
            For i = 0 To bestMultiPos.Length - 1
                s &= bestMultiPos(i).ToString("F2") & " "
            Next i
            s &= "] "
            s &= "Best Multiswarm Cost = " & bestMultiCost.ToString("F3")
            s &= vbCrLf & "=======================" & vbCrLf
            Return s
        End Function

    End Class ' Multiswarm

End Namespace ' ns


